module TreeZipper where

import Prelude hiding (zip)

data Cxt a = Top | L (Cxt a) a | R a (Cxt a)
    deriving (Eq, Ord, Show) 

data Tree = Leaf Int | Fork Tree Tree
     deriving (Eq, Ord, Show)

type Loc a = (a, Cxt a)

left :: Loc Tree -> Loc Tree
left (Fork l r, c) = (l, L c r)

right :: Loc Tree -> Loc Tree
right (Fork l r, c) = (r, R l c)

isRoot (_, Top) = True
isRoot _ = False

parent :: Loc Tree -> Loc Tree
parent (t, L c r) = (Fork t r, c)
parent (t, R l c) = (Fork l t, c)

top :: Tree -> Loc Tree
top t = (t, Top)

value :: Loc Tree -> Tree
value = fst

zip :: Tree -> Loc Tree
zip = top

tree = Fork (Leaf 1)
            (Fork (Leaf 4)
                  (Leaf 7))

subtree = left (right (zip tree))

edt = let (Leaf v, cxt) = subtree
          subtree' = (Leaf (v+1), cxt)
      in value (parent (parent subtree'))


