{-# LANGUAGE GADTs, EmptyDataDecls, ExistentialQuantification, ScopedTypeVariables, MultiParamTypeClasses, FunctionalDependencies, FlexibleContexts, FlexibleInstances, UndecidableInstances #-}

module Zipper where

import Data.Data

data Zipper l h r up =
    Zipper h (Context l h r up)

data Context l h r up where
    ContTop :: Context Null h Null Null
    Cont    :: Left l (h -> r)
            -> Right r h_parent
            -> Context l_parent h_parent r_parent up
            -> Context l h r (Up l_parent h_parent r_parent up)

data Up l a r up

data Null  

data Left contains expects where
    LeftUnit :: a -> Left Null a
    LeftCons :: Left c (b -> a) -> b -> Left (c -> b) a

data Right provides final where
    RightNull :: Right final final
    RightCons :: b -> Right a final -> Right (b -> a) final


{- Zipper operations -}

zip :: hole -> Zipper Null hole Null Null
zip a = Zipper a ContTop

value :: Zipper l h r up -> h
value (Zipper h _) = h

set :: h -> Zipper l h r up -> Zipper l h r up
set h (Zipper _ context) = Zipper h context

move_left :: Zipper (l -> h_new) h_old r up
          -> Zipper l h_new (h_old -> r) up
move_left (Zipper h_old (Cont (LeftCons l h_new) r up))
          = Zipper h_new (Cont l (RightCons h_old r) up)

move_right :: Zipper l h_old (h_new -> r) up
           -> Zipper (l -> h_old) h_new r up
move_right (Zipper h_old (Cont l (RightCons h_new r) up))
           = Zipper h_new (Cont (LeftCons l h_old) r up)


parent :: Zipper l h r (Up l' h' r' parent)
           -> Zipper l' h' r' parent
parent (Zipper h (Cont l r up))
           = Zipper (collapse l h r) 
                    up

collapse :: Left l (h -> r) -> h -> Right r final -> final
collapse l h r = total where
    left = collapse_left l
    mid = left h
    total = collapse_right mid r

collapse_left :: Left l r -> r
collapse_left (LeftUnit a) = a
collapse_left (LeftCons f b) = collapse_left f b

collapse_right :: r -> Right r final -> final
collapse_right f RightNull = f
collapse_right f (RightCons b r) = collapse_right (f b) r


child :: (Typeable l_down, Typeable h_down, Typeable l, Data h, Typeable r, Typeable up, Foldl Null t h (l_down -> h_down)) 
              => Zipper l h r up -> t -> Zipper l_down h_down h (Up l h r up)
child z _ = child_aux z

child_aux :: forall h l_down h_down l r up. (Typeable l_down, Typeable h_down, Typeable l, Data h, Typeable r, Typeable up)
             => Zipper l h r up
             -> Zipper l_down h_down h (Up l h r up)
child_aux (Zipper h c)
          = case (gfoldl erased_left_cons erased_left_unit h) of
              Erase l -> case cast l of
                  Just ((LeftCons l' h_down) :: Left (l_down -> h_down) h)
                               -> Zipper h_down (Cont l' RightNull c)

data Erase c a = forall b. (Typeable b) => Erase (c b a)

erased_left_cons :: (Typeable b) 
                    => Erase Left (b -> a) -> b -> Erase Left a
erased_left_cons (Erase c) b = Erase (LeftCons c b)

erased_left_unit :: a -> Erase Left a
erased_left_unit a = Erase (LeftUnit a)

instance Typeable Null where
    typeOf _ = mkTyConApp (mkTyCon "Null") []

instance Typeable2 Left where
    typeOf2 _ = mkTyConApp (mkTyCon "Left") []

instance Typeable4 Up where
    typeOf4 _ =  mkTyConApp (mkTyCon "Up") []

class Foldl acc right stop left
    | acc right stop -> left where

instance Foldl acc stop stop acc where

instance Foldl (acc -> a) b stop left => Foldl acc (a->b) stop left where
